"""
TODO:
    - look into https://github.com/click-contrib/click-help-colors
"""

from csv import DictReader
import datetime
from distutils.util import strtobool
try:
    from importlib.metadata import version
except ImportError:
    from importlib_metadata import version
import json
import re
import sys
from typing import Dict, Generator, Iterable, List, Optional, Tuple, Union
import urllib.parse
import uuid

import click
from globus_sdk import TransferClient

from timer_cli.auth import get_current_user
from timer_cli.auth import logout as auth_logout
from timer_cli.auth import revoke_login
from timer_cli.job import (
    job_delete,
    job_list,
    job_status,
    job_submit,
    show_job,
    show_job_list,
)
from timer_cli.output import make_table
from timer_cli.transfer import (
    TRANSFER_ALL_SCOPE,
    error_if_not_activated,
    get_transfer_client,
)

# List of datetime formats accepted as input. (`%z` means timezone.)
DATETIME_FORMATS = [
    "%Y-%m-%d",
    "%Y-%m-%dT%H:%M:%S",
    "%Y-%m-%dT%H:%M:%S%z",
    "%Y-%m-%d %H:%M:%S",
    "%Y-%m-%d %H:%M:%S%z",
]


# TODO: make configurable/environment/maybe even check terminal width?
MAX_CONTENT_WIDTH = 100


timedelta_regex = re.compile(
    r"\s*((?P<weeks>\d+)w)?"
    r"\s*((?P<days>\d+)d)?"
    r"\s*((?P<hours>\d+)h)?"
    r"\s*((?P<minutes>\d+)m)?"
    r"\s*((?P<seconds>\d+)s?)?"
)


INTERVAL_HELP = (
    "Interval at which the job should run. Use 'w', 'd', 'h', 'm', and 's' as suffixes"
    " to specify weeks, days, hours, minutes, and seconds. Examples: '1h 30m', '500s',"
    " '24h', '1d 12h', '2w', etc. Must be in order: hours -> minutes -> seconds. You"
    " should either use quotes ('1d 2h') or write without spaces (1d2h)."
)


def _parse_timedelta(s: str) -> datetime.timedelta:
    groups = {k: int(v) for k, v in timedelta_regex.match(s).groupdict(0).items()}
    # timedelta accepts kwargs for units up through days, have to convert weeks
    groups["days"] += groups.pop("weeks", 0) * 7
    return datetime.timedelta(**groups)


def _un_parse_opt(opt: str):
    return "--" + opt.replace("_", "-")


def _read_csv(
    file_name: str,
    fieldnames=["source_path", "destination_path", "recursive"],
    comment_char: str = "#",
) -> Generator[Dict[str, Union[str, bool]], None, None]:
    def decomment(f):
        for row in f:
            if not row.startswith(comment_char):
                yield row

    def transform_val(k: str, v: str) -> Union[str, bool]:
        v = v.strip()
        # Was hoping to make this a bit more generic, but spent enough time on it so,
        # handling the case we actually need here
        if k == "recursive":
            try:
                return bool(strtobool(v))
            except ValueError:
                # "invalid truth value"
                click.echo(f"In file {file_name}: couldn't parse {v} as a truth value")
                sys.exit(1)
        else:
            return v

    with open(file_name, "r") as f:
        reader = DictReader(decomment(f), fieldnames=fieldnames)
        for row_dict in reader:
            yield {k: transform_val(k, v) for k, v in row_dict.items()}


def _get_required_data_access_scopes(
    tc: TransferClient,
    collection_ids: Iterable[str],
) -> List[str]:
    data_access_scopes: List[str] = []
    for collection_id in collection_ids:
        collection_id_info = tc.get_endpoint(collection_id)
        if collection_id_info["DATA_TYPE"] == "endpoint":
            gcs_version = collection_id_info.get("gcs_version")
            if gcs_version is None:
                continue
            gcs_version_parts = [int(x) for x in gcs_version.split(".")]
            requires_data_access = all(
                [
                    gcs_version_parts[0] > 5
                    or gcs_version_parts[0] == 5
                    and gcs_version_parts[1] >= 4,
                    (collection_id_info.get("high_assurance", True) is False),
                    collection_id_info.get("host_endpoint", True) is None,
                ]
            )
            if requires_data_access:
                data_access_scopes.append(
                    f"https://auth.globus.org/scopes/{collection_id}/data_access"
                )
    return data_access_scopes


def get_version_string() -> str:
    v = version("globus-timer-cli")
    return f"globus-timer-cli {v}"


def show_version_string():
    click.echo(get_version_string())


def show_usage(cmd: click.Command):
    """
    Show the relevant usage and exit.

    The actual usage message is also specific to incomplete commands, thanks to using
    the click context.
    """
    ctx = click.get_current_context()
    # TODO: disabling this next line for the time being because of inconsistent
    # behavior between this function and calling --help directly, which would produce
    # different output. still have to figure that out
    # ctx.max_content_width = MAX_CONTENT_WIDTH
    formatter = ctx.make_formatter()
    cmd.format_help_text(ctx, formatter)
    cmd.format_options(ctx, formatter)
    cmd.format_epilog(ctx, formatter)
    click.echo(formatter.getvalue().rstrip("\n"))
    ctx.exit(2)


class Command(click.Command):
    """
    Subclass click.Command to show help message if a command is missing arguments.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def make_context(self, *args, **kwargs):
        """
        This is a semi-internal method on Commands that starts the parsing to create a
        new context. We hook into that here to catch any immediate parsing errors
        (missing arguments etc.) to exit and show our usage message.
        """
        try:
            return super().make_context(*args, **kwargs)
        except (
            click.BadArgumentUsage,
            click.BadOptionUsage,
            click.BadParameter,
            click.MissingParameter,
        ) as e:
            e.cmd = None
            e.show()
            click.echo()
            show_usage(self)


class MutuallyExclusive(click.Option):
    """
    Based on the answer for: https://stackoverflow.com/q/44247099
    """

    def __init__(self, *args, **kwargs):
        self.mutually_exclusive = kwargs.pop("mutually_exclusive")
        assert self.mutually_exclusive, "'mutually_exclusive' parameter required"
        super().__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        self_exists = self.name in opts
        mutually_exclusive_exists = self.mutually_exclusive in opts
        if self_exists and mutually_exclusive_exists:
            raise click.UsageError(
                f"Illegal usage: `{self.name}` is mutually exclusive with"
                f" `{self.mutually_exclusive}`"
            )
        if not (self_exists or mutually_exclusive_exists):
            full_options_names = ", ".join(
                [
                    "/".join(opt.opts)
                    for opt in ctx.command.params
                    if opt.name in self.mutually_exclusive or opt == self
                ]
            )
            raise click.UsageError(
                "Illegal usage: one of the following options is required:"
                f" {full_options_names}"
            )
        return super().handle_parse_result(ctx, opts, args)


def _get_options_flags(options: Iterable[click.Option]):
    """
    Given a list of options, produce a list of formatted option flags, like
    "--verbose/-v".
    """
    return ", ".join(["/".join(opt.opts) for opt in options])


class JointlyExhaustive(click.Option):
    """
    "Jointly exhaustive", meaning that at least one must occur.
    """

    def __init__(self, *args, **kwargs):
        self.jointly_exhaustive: List[str] = kwargs.pop("jointly_exhaustive")
        assert self.jointly_exhaustive, "'jointly_exhaustive' parameter required"
        super().__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        options_exist = [opt in opts for opt in self.jointly_exhaustive]
        options_exist.append(self.name in opts)
        if not any(options_exist):
            ctx = click.get_current_context()
            full_options_names = ", ".join(
                [
                    "/".join(opt.opts)
                    for opt in ctx.command.params
                    if opt.name in self.jointly_exhaustive or opt == self
                ]
            )
            raise click.UsageError(
                "Must provide at least one of the following options:"
                f" {full_options_names}"
            )
        return super().handle_parse_result(ctx, opts, args)


class URL(click.ParamType):
    """Click param type for a URL."""

    name = "url"

    def convert(self, value, param, ctx):
        if not isinstance(value, tuple):
            value = urllib.parse.urlparse(value)
            if not value.netloc:
                self.fail("incomplete URL")
            if value.scheme not in ("http", "https"):
                self.fail(
                    f"invalid URL scheme ({value.scheme}). Only HTTP URLs are allowed",
                    param,
                    ctx,
                )
        return value


@click.group(invoke_without_command=True, no_args_is_help=True)
@click.option(
    "--version",
    "-v",
    "show_version",
    is_flag=True,
    default=False,
    help="Print what version of this CLI is installed."
)
def cli(show_version: bool):
    if show_version:
        show_version_string()
        return


@cli.group(help="Commands for managing periodic Globus Transfer jobs.")
def job():
    pass


@job.command(cls=Command)
@click.option(
    "--name",
    required=True,
    type=str,
    help="Name to identify this job in the timer service (not necessarily unique)",
)
@click.option(
    "--start",
    required=False,
    type=click.DateTime(formats=DATETIME_FORMATS),
    help=(
        "Start time for the job. Defaults to current time. (The example above shows the"
        " allowed formats using Python's datetime formatters; see:"
        " https://docs.python.org/3/library/datetime.html"
        "#strftime-and-strptime-format-codes"
    ),
)
@click.option(
    "--interval",
    required=True,
    type=str,
    help=INTERVAL_HELP,
)
@click.option(
    "--scope",
    required=True,
    type=str,
    help="Globus Auth scope needed for this action",
)
@click.option(
    "--action-url",
    required=True,
    type=URL(),
    help=(
        "The URL for the action to run, e.g. "
        "https://actions.automate.globus.org/transfer/transfer/run"
    ),
)
@click.option(
    "--action-body",
    required=False,
    type=str,
    help=(
        "request JSON body to send to the action provider on job execution (NOTE:"
        " mutually exclusive with --action-file)"
    ),
    cls=MutuallyExclusive,
    mutually_exclusive="action_file",
)
@click.option(
    "--action-file",
    required=False,
    type=click.File("r"),
    help=(
        "path to a file containing JSON to send to the action provider on job"
        " execution (NOTE: mutually exclusive with --action-body)"
    ),
    cls=MutuallyExclusive,
    mutually_exclusive="action_body",
)
@click.option(
    "--verbose",
    "-v",
    is_flag=True,
    help="Show full JSON output",
)
def submit(
    name: str,
    start: Optional[datetime.datetime],
    interval: str,
    scope: str,
    action_url: urllib.parse.ParseResult,
    action_body: Optional[str],
    action_file: Optional[click.File],
    verbose: bool,
):
    """
    Submit a job.
    """
    interval_seconds = _parse_timedelta(interval).total_seconds()
    if not interval_seconds:
        raise click.UsageError(f"Couldn't parse interval: {interval}")
    if interval_seconds < 60:
        raise click.UsageError(f"Interval is too short, minimum is 1 minute")
    response = job_submit(
        name,
        start,
        interval_seconds,
        scope,
        action_url,
        action_body,
        action_file,
    )
    show_job(response, verbose=verbose)


@job.command(cls=Command)
@click.option(
    "--show-deleted",
    required=False,
    is_flag=True,
    help="Whether to include deleted jobs in the output",
)
@click.option(
    "--verbose",
    "-v",
    is_flag=True,
    help="Show full JSON output",
)
def list(show_deleted: bool, verbose: bool):
    """
    List submitted jobs.

    Note that, in the non-verbose output, `Last Result` is reporting according to
    whether Automate could successfully submit the job. It's possible for Transfer
    to run into errors attempting to run your submission, which timer/Automate are not
    aware of.

    CHECK THE --verbose OUTPUT TO BE CERTAIN YOUR TRANSFERS ARE WORKING.
    """
    response = job_list(show_deleted=show_deleted)
    show_job_list(response, verbose=verbose)


@job.command(cls=Command)
@click.option(
    "--all",
    "-a",
    "show_all",
    is_flag=True,
    help="Show status for all jobs",
)
@click.option(
    "--show-deleted",
    required=False,
    is_flag=True,
    help="Whether to include deleted jobs in the output",
)
@click.option(
    "--verbose",
    "-v",
    is_flag=True,
    help="Show full JSON output",
)
@click.argument("job_id", type=uuid.UUID, required=False)
def status(
    job_id: Optional[uuid.UUID], show_deleted: bool, verbose: bool, show_all: bool
):
    """
    Return the status of the job with the given ID.

    Note that, in the non-verbose output, `Last Result` is reporting according to
    whether Automate could successfully submit the job. It's possible for Transfer
    to run into errors attempting to run your submission, which timer/Automate are not
    aware of.

    CHECK THE --verbose OUTPUT TO BE CERTAIN YOUR TRANSFERS ARE WORKING.
    """
    if not job_id and not show_all:
        click.echo(
            "Error: must provide either a job ID or the --all option\n", err=True
        )
        show_usage(click.get_current_context().command)
    if show_all:
        show_job_list(job_list(), as_table=False, verbose=verbose)
    else:
        show_job(job_status(job_id, show_deleted=show_deleted), verbose=verbose)


@job.command(cls=Command)
@click.option(
    "--verbose",
    "-v",
    is_flag=True,
    help="Show full JSON output",
)
@click.argument("job_ids", type=uuid.UUID, nargs=-1)
def delete(job_ids: Iterable[uuid.UUID], verbose: bool):
    start = True
    for job_id in job_ids:
        if not start:
            click.echo("")
        show_job(job_delete(job_id), verbose=verbose, was_deleted=True)
        start = False


@job.command(cls=Command)
@click.option(
    "--name",
    required=True,
    type=str,
    help="Name to identify this job in the timer service (not necessarily unique)",
)
@click.option(
    "--start",
    required=False,
    type=click.DateTime(formats=DATETIME_FORMATS),
    help=("Start time for the job (defaults to current time)"),
)
@click.option(
    "--interval",
    required=True,
    type=str,
    help=INTERVAL_HELP,
)
@click.option(
    "--source-endpoint",
    required=True,
    type=str,
    help="ID for the source transfer endpoint",
)
@click.option(
    "--dest-endpoint",
    required=True,
    type=str,
    help="ID for the destination transfer endpoint",
)
@click.option(
    "--label",
    required=False,
    type=str,
    help=(
        "An optional label for the transfer operation, up to 128 characters long. Must"
        " contain only letters/numbers/spaces, and the following characters: - _ ,"
    ),
)
@click.option(
    "--stop-after-date",
    required=False,
    type=click.DateTime(formats=DATETIME_FORMATS),
    help=("Stop running the transfer after this date"),
)
@click.option(
    "--stop-after-runs",
    required=False,
    type=int,
    help=("Stop running the transfer after this number of runs have happened"),
)
@click.option(
    "--sync-level",
    required=False,
    type=int,
    help=(
        "Specify that only new or modified files should be transferred. The behavior"
        " depends on the value of this parameter, which must be a value 0--3, as"
        " defined in the transfer API: 0. Copy files that do not exist at the"
        " destination. 1. Copy files if the size of the destination does not match the"
        " size of the source. 2. Copy files if the timestamp of the destination is"
        " older than the timestamp of the source. 3. Copy files if checksums of the"
        " source and destination do not match."
    ),
)
@click.option(
    "--encrypt-data",
    is_flag=True,
    default=False,
    help="Whether Transfer should encrypt data sent through the network using TLS",
)
@click.option(
    "--verify-checksum",
    is_flag=True,
    default=False,
    help=(
        "Whether Transfer should verify file checksums and retry if the source and"
        " destination don't match"
    ),
)
@click.option(
    "--preserve-timestamp",
    is_flag=True,
    default=False,
    help=(
        "Whether Transfer should set file timestamps on the destination to match the"
        " origin"
    ),
)
@click.option(
    "--item",
    "-i",
    required=False,
    type=(str, str, bool),
    multiple=True,
    cls=JointlyExhaustive,
    jointly_exhaustive=["items_file"],
    help=(
        "Used to specify the transfer items; provide as many of this option as files"
        " to transfer. The format for this option is `--item SRC DST RECURSIVE`, where"
        " RECURSIVE specifies, if this item is a directory, to transfer the entire"
        " directory. For example: `--item ~/file1.txt ~/new_file1.txt false`"
    ),
)
@click.option(
    "--items-file",
    required=False,
    type=str,
    cls=JointlyExhaustive,
    jointly_exhaustive=["item"],
    help="file containing table of items to transfer",
)
@click.option(
    "--verbose",
    "-v",
    is_flag=True,
    default=False,
    help="Show full JSON output",
)
def transfer(
    name: str,
    start: Optional[datetime.datetime],
    interval: str,
    source_endpoint: str,
    dest_endpoint: str,
    label: Optional[str],
    stop_after_date: Optional[datetime.datetime],
    stop_after_runs: Optional[int],
    sync_level: Optional[int],
    encrypt_data: bool,
    verify_checksum: bool,
    preserve_timestamp: bool,
    item: Optional[List[Tuple[str, str, Optional[str]]]],
    items_file: Optional[str],
    verbose: bool,
):
    """
    Submit a task for periodic transfer or sync using Globus transfer. The options for
    this command are tailored to the transfer action.
    """
    action_url = urllib.parse.urlparse(
        "https://actions.automate.globus.org/transfer/transfer/run"
    )
    endpoints = [source_endpoint, dest_endpoint]
    tc = get_transfer_client()
    error_if_not_activated(tc, endpoints)
    data_access_scopes = _get_required_data_access_scopes(tc, endpoints)
    transfer_ap_scope = (
        "https://auth.globus.org/scopes/actions.globus.org/transfer/transfer"
    )
    if len(data_access_scopes) > 0:
        transfer_ap_scope = (
            f"{transfer_ap_scope}[{TRANSFER_ALL_SCOPE}[{' '.join(data_access_scopes)}]]"
        )
    # Just declare it for typing purposes
    transfer_items: List[Dict[str, Union[str, bool]]] = []
    if item:
        transfer_items = [
            {
                "source_path": i[0].strip(),
                "destination_path": i[1].strip(),
                "recursive": i[2],
            }
            for i in item
        ]
    elif items_file:
        # Unwind the generator
        transfer_items = [i for i in _read_csv(items_file)]

    action_body = {
        "source_endpoint_id": source_endpoint,
        "destination_endpoint_id": dest_endpoint,
        "transfer_items": transfer_items,
    }
    if label:
        action_body["label"] = label
    else:
        action_body["label"] = f"Job from Timer service named {name}"
    if sync_level is not None:
        action_body["sync_level"] = sync_level
    action_body["encrypt_data"] = encrypt_data
    action_body["verify_checksum"] = verify_checksum
    action_body["preserve_timestamp"] = preserve_timestamp
    callback_body = {"body": action_body}
    interval_seconds = _parse_timedelta(interval).total_seconds()
    if not interval_seconds:
        raise click.UsageError(f"Couldn't parse interval: {interval}")
    if interval_seconds < 60:
        raise click.UsageError(f"Interval is too short, minimum is 1 minute")
    response = job_submit(
        name,
        start,
        interval_seconds,
        transfer_ap_scope,
        action_url,
        action_body=None,
        action_file=None,
        callback_body=callback_body,
        stop_after_date=stop_after_date,
        stop_after_runs=stop_after_runs,
    )
    show_job(response, verbose=verbose)


@cli.group(help="Commands related to managing your Globus Auth credentials.")
def session():
    pass


@session.command(
    help="Cache identity information for future operations. This is "
    "optional, as it will be done on demand if this command is not used."
)
def login():
    user_info = get_current_user()
    click.echo(f"Logged in as {user_info['preferred_username']}")


@session.command(help="Print information about your currently logged in session.")
@click.option(
    "--format",
    type=click.Choice(["brief", "full", "json"], case_sensitive=False),
    default="brief",
    help="Select the detail level and format for output.",
)
def whoami(format: str):
    user_info = get_current_user(no_login=True)
    if not user_info:
        click.echo(
            "Not logged in yet; use `globus-timer session login` to initialize your"
            " session",
            err=True,
        )
        sys.exit(1)
    full_fields = ["name", "email", "preferred_username", "organization"]
    if format == "brief":
        click.echo(f"{user_info['preferred_username']}")
    elif format == "full":
        click.echo(make_table(full_fields, [[user_info[k] for k in full_fields]]))
    elif format == "json":
        click.echo(json.dumps({k: user_info[k] for k in full_fields}, indent=2))


@session.command(help="Remove the saved Globus Auth identity identity information.")
def logout():
    logged_out = auth_logout()
    if logged_out:
        click.echo("Successfully logged out.")
    else:
        click.echo("Unable to remove stored tokens to perform the logout.")


@session.command(help="Remove Timer's authorization to use your credentials.")
def revoke():
    if revoke_login():
        click.echo("Successfully revoked permission for all Timer operations.")
    else:
        click.echo("Unable to revoke login.")


def main():
    cli()


if __name__ == "__main__":
    cli()



import datetime
import json
import os
import sys
from typing import Callable, Dict, Iterable, List, Optional, Tuple, Union
import urllib
import uuid

import click
import requests

from timer_cli.auth import TIMER_SERVICE_SCOPE, get_access_token_for_scope
from timer_cli.output import make_table, show_response

# how long to wait before giving up on requests to the API
TIMEOUT = 10


_DEFAULT_TIMER_SERVICE_URL = "https://timer.automate.globus.org"

TIMER_SERVICE_URL = os.environ.get("TIMER_SERVICE_URL", _DEFAULT_TIMER_SERVICE_URL)

_TIMER_JOBS_URL = f"{TIMER_SERVICE_URL}/jobs"


def handle_requests_exception(e: Exception):
    click.echo(f"error in request: {e}", err=True)
    sys.exit(1)


def get_headers(token_scope: str = TIMER_SERVICE_SCOPE) -> Dict[str, str]:
    """
    Assemble any needed headers that should go in all requests to the timer API, such
    as the access token.
    """
    token = get_access_token_for_scope(token_scope)
    if not token:
        raise ValueError("couldn't get token")
    return {"Authorization": f"Bearer {token}"}


def job_submit(
    name: str,
    start: Optional[datetime.datetime],
    interval: int,
    scope: Optional[str],
    action_url: urllib.parse.ParseResult,
    action_body: Optional[str] = None,
    action_file: Optional[click.File] = None,
    callback_body: Optional[dict] = None,
    stop_after_date: Optional[datetime.datetime] = None,
    stop_after_runs: Optional[int] = None,
) -> requests.Response:
    if not callback_body:
        try:
            if action_body:
                action_body = action_body.strip("'").strip('"')
                callback_body = json.loads(action_body)
            elif action_file is not None:  # action_file
                callback_body = json.load(action_file)
        except (TypeError, ValueError) as e:
            raise click.BadOptionUsage(
                "action-body",
                f"--action-body must parse into valid JSON; got error: {e}",
            )
    start = start or datetime.datetime.now()
    start_with_tz = start
    if start_with_tz.tzinfo is None:
        start_with_tz = start.astimezone()
    callback_url: str = action_url.geturl()
    # If there is a dependent scope on the job's scope, we trim that off as server-side
    # doesn't really understand that.
    send_job_scope = scope.split("[", 1)[0]
    req_json = {
        "name": name,
        "start": start_with_tz.isoformat(),
        "interval": interval,
        "scope": send_job_scope,
        "callback_url": callback_url,
        "callback_body": callback_body,
    }
    if stop_after_date or stop_after_runs:
        req_json["stop_after"] = dict()
        if stop_after_date:
            req_json["stop_after"]["date"] = stop_after_date
        if stop_after_runs:
            req_json["stop_after"]["n_runs"] = stop_after_runs
    # Ww'll make the job scope dependent on the timer service's job creation scope so we
    # can be sure that we can do a dependent token grant on the token that is sent
    token_scope = f"{TIMER_SERVICE_SCOPE}[{scope}]"
    try:
        headers = get_headers(token_scope=token_scope)
    except (EnvironmentError, ValueError) as e:
        click.echo(str(e), err=True)
        sys.exit(1)
    try:
        return requests.post(
            _TIMER_JOBS_URL,
            json=req_json,
            headers=headers,
            timeout=TIMEOUT,
        )
    except requests.RequestException as e:
        handle_requests_exception(e)


def job_list(show_deleted: bool = False) -> requests.Response:
    headers = get_headers()
    params = dict()
    if show_deleted:
        params["show_deleted"] = True
    try:
        return requests.get(
            f"{_TIMER_JOBS_URL}",
            params=params,
            headers=headers,
            timeout=TIMEOUT,
        )
    except requests.RequestException as e:
        handle_requests_exception(e)


def job_status(job_id: uuid.UUID, show_deleted: bool = False) -> requests.Response:
    headers = get_headers()
    params = dict()
    if show_deleted:
        params["show_deleted"] = True
    try:
        return requests.get(
            f"{_TIMER_JOBS_URL}/{job_id}",
            params=params,
            headers=headers,
            timeout=TIMEOUT,
        )
    except requests.RequestException as e:
        handle_requests_exception(e)


def job_delete(job_id: uuid.UUID) -> requests.Response:
    headers = get_headers()
    try:
        return requests.delete(
            f"{_TIMER_JOBS_URL}/{job_id}",
            headers=headers,
            timeout=TIMEOUT,
        )
    except requests.RequestException as e:
        handle_requests_exception(e)


def _get_job_result(job_json: dict) -> Optional[str]:
    if "results" not in job_json:
        return None
    last_result = "RUN COMPLETE"
    job_results = job_json["results"]
    if len(job_results) == 0:
        last_result = "NOT RUN"
    else:
        last_result_status = job_results[0].get("status") or 500
        # this could be better
        if last_result_status >= 400:
            last_result = "FAILURE"
    return last_result


def _job_prop_name_map(
    job_json: Dict, prop_map: List[Tuple[str, Union[str, Callable]]]
) -> List[Tuple[str, str]]:
    """
    Map the job json (or any dict) from some unfriendly names to more friendly names on
    the keys. The input is the dict, and a list of friendly names, followed by
    unfriendly names or a callable which will generate the desired value for the
    friendly name.

    The return is a list of friendly name, value tuples
    """
    ret_map: List[Tuple[str, str]] = []
    for prop_name, prop_mapper in prop_map:
        value = job_json.get(prop_mapper)
        if prop_mapper not in job_json and callable(prop_mapper):
            value = prop_mapper(job_json)
        ret_map.append((prop_name, value))
    return ret_map


def show_job(response: requests.Response, verbose: bool, was_deleted: bool = False):
    if response.status_code >= 300:
        try:
            msg = response.json().get("error", dict()).get("detail")
        except ValueError:
            msg = None
        finally:
            if msg:
                msg = f": {msg}"
            else:
                msg = ""
        click.echo(
            f"Unable to retrieve job{msg}\n"
            f"    Response status:  {response.status_code}\n"
            f"    Response body:    {response.text}"
        )
        return
    if verbose:
        return show_response(response)
    try:
        job_json = response.json()
    except ValueError as e:
        # TODO: this could return status---maybe add module with error hierarchy
        click.echo(f"couldn't parse json from job response: {e}", err=True)
        return
    show_job_json(job_json, was_deleted=was_deleted)


def show_job_json(job_json: dict, was_deleted: bool = False):
    try:
        job_friendly_to_field_map = [
            ("Name", "name"),
            ("Job ID", "job_id"),
            ("Status", "status"),
            ("Start", "start"),
            ("Interval", lambda d: str(datetime.timedelta(seconds=d["interval"]))),
        ]
        if not was_deleted:
            job_friendly_to_field_map.append(("Next Run At", "next_run"))
            job_friendly_to_field_map.append(("Last Run Result", _get_job_result))
    except (IndexError, KeyError) as e:
        click.echo(f"failed to read info for job: {str(e)}", err=True)
        return
    job_info = _job_prop_name_map(job_json, job_friendly_to_field_map)
    key_width = max(len(k) for k, _ in job_info) + 2
    output = "\n".join([f"{k}: ".ljust(key_width) + str(v) for k, v in job_info])
    click.echo(output)


def show_job_list(
    response: requests.Response,
    verbose: bool = False,
    as_table: bool = True,
) -> None:
    # TODO: absorb this chunk into shared function
    if verbose:
        return show_response(response)
    try:
        job_json = response.json()
    except ValueError as e:
        # TODO: bad exit, do errors
        click.echo(
            f"couldn't parse json from job response: {e}: service response: {response.text}",
            err=True,
        )
        sys.exit(1)

    if not as_table:
        if "jobs" not in job_json:
            click.echo(f"failed to read info for job list: {e}", err=True)
            sys.exit(1)
        first = True
        for job in job_json["jobs"]:  # are we semantically satiated yet?
            # print empty separating line for each job after the first
            if not first:
                click.echo("")
            show_job_json(job)
            first = False
        return

    headers = ["Name", "Job ID", "Status", "Last Result"]
    try:
        rows = [
            [job["name"], job["job_id"], job["status"], _get_job_result(job)]
            for job in job_json["jobs"]
        ]
    except (IndexError, KeyError) as e:
        # TODO: bad exit, do errors
        click.echo(f"failed to read info for job: {e}", err=True)
        sys.exit(1)
    table = make_table(headers, rows)
    click.echo(table)



import sys
from typing import List

import click
from globus_sdk import GlobusError, TransferClient

from timer_cli.auth import get_authorizers_for_scopes

TRANSFER_ALL_SCOPE = "urn:globus:auth:scope:transfer.api.globus.org:all"


def endpoints_not_activated(
    transfer_client: TransferClient, endpoints: List[str]
) -> List[str]:
    """
    Filter list of endpoint UUIDs, returning unactivated ones.

    Exit 1 if transfer responds with an error trying to look up endpoints.
    """
    result = []
    for endpoint in endpoints:
        try:
            if not transfer_client.get_endpoint(endpoint).get("activated"):
                result.append(endpoint)
        except GlobusError as e:
            click.echo(
                f"couldn't get information for endpoint {endpoint}:"
                f" {e.code}, {e.message}",
                err=True,
            )
            sys.exit(1)
    return result


def error_if_not_activated(
    transfer_client: TransferClient,
    endpoints: List[str],
    reactivate_if_expires_in=86400,
):
    not_activated = endpoints_not_activated(transfer_client, endpoints)
    still_not_activated = []
    for endpoint in not_activated:
        response = transfer_client.endpoint_autoactivate(
            endpoint, if_expires_in=reactivate_if_expires_in
        )
        if response.get("code") == "AutoActivationFailed":
            still_not_activated.append(endpoint)
    if still_not_activated:
        click.echo(
            f"Error: requested endpoint is not activated: {', '.join(not_activated)}\n"
            "Open in the web app to activate:",
            err=True,
        )
        for endpoint in not_activated:
            click.echo(
                f"    https://app.globus.org/file-manager?origin_id={endpoint}",
                err=True,
            )
        sys.exit(1)


def get_transfer_client():
    authorizers = get_authorizers_for_scopes([TRANSFER_ALL_SCOPE])
    return TransferClient(authorizers.get(TRANSFER_ALL_SCOPE))



import json
from typing import List

import click
import requests


def make_table(headers: List[str], rows: List[List[str]]) -> str:
    contents = [headers] + rows
    widths = [max(len(cell) for cell in col) for col in zip(*contents)]
    formatted_header = " | ".join(
        cell.ljust(width) for cell, width in zip(headers, widths)
    )
    formatted_rows = [
        " | ".join(cell.ljust(width) for cell, width in zip(row, widths))
        for row in rows
    ]
    max_row_width = len(formatted_header)
    if formatted_rows:
        max_row_width = max(max_row_width, max(map(len, formatted_rows)))
    separator = "-|-".join("-" * w for w in widths)
    formatted_rows_block = "\n".join(formatted_rows)
    return f"{formatted_header}\n{separator}\n{formatted_rows_block}"


def show_response(response: requests.Response):
    if response.status_code >= 400:
        click.echo(f"got response code {response.status_code}", err=True)
    click.echo(json.dumps(response.json(), indent=2))



from collections import defaultdict
import json
from json import JSONDecodeError
import os
import pathlib
import platform
import sys
from typing import Any, Dict, List, NamedTuple, Optional, Set

import click
from globus_sdk import AuthClient, NativeAppAuthClient
from globus_sdk.auth.token_response import OAuthTokenResponse
from globus_sdk.authorizers import GlobusAuthorizer, RefreshTokenAuthorizer
from globus_sdk.exc import AuthAPIError

CLIENT_ID = "bc77d044-1f42-46cc-9702-87f756cd08a6"
CLIENT_NAME = "Globus Timer Command Line Interface"
TIMER_SERVICE_SCOPE = "https://auth.globus.org/scopes/524230d7-ea86-4a52-8312-86065a9e0417/transfer_action"
AUTH_SCOPES = [
    "openid",
    "email",
    "profile",
]
ALL_SCOPES = AUTH_SCOPES + [TIMER_SERVICE_SCOPE]

DEFAULT_TOKEN_FILE = pathlib.Path.home() / pathlib.Path(".globus_timer_tokens.json")


def _get_base_scope(scope: str):
    if "[" in scope:
        return scope.split("[")[0]
    return scope


class TokenSet(NamedTuple):
    """
    Might want to check out this as a replacement:
    https://www.attrs.org/en/stable/why.html#namedtuples
    """

    access_token: str
    refresh_token: Optional[str]
    expiration_time: Optional[int]
    # Keep track of scopes associated with these tokens with the dependencies still
    # included. If we need to get a token where tokens for the base scope exist but
    # there isn't a matching dependent scope, that means we need to prompt for consent
    # again. If there is a matching full-scope-string in `dependent_scopes`, then we're
    # OK to use the token from looking up that base scope.
    dependent_scopes: Set[str]


class TokenCache:
    def __init__(self, token_store: str):
        self.token_store = token_store
        self.tokens: Dict[str, TokenSet] = {}
        self.modified = False

    def set_tokens(self, scope: str, tokens: TokenSet) -> TokenSet:
        if scope in self.tokens:
            dependent_scopes = set(tokens.dependent_scopes).union(
                set(self.tokens[scope].dependent_scopes)
            )
            new_token_set = TokenSet(
                access_token=tokens.access_token,
                refresh_token=tokens.refresh_token,
                expiration_time=tokens.expiration_time,
                dependent_scopes=dependent_scopes,
            )
            self.tokens[scope] = new_token_set
        else:
            self.tokens[scope] = tokens
        self.modified = True
        return tokens

    def get_tokens(self, scope: str) -> Optional[TokenSet]:
        if "[" in scope:
            # If the full scope string is in our mapping already, we can just return
            # the tokens. If not, even if we have a token for the base scope, we
            # shouldn't return that because it won't work for the new scope.
            base_scope = scope.split("[")[0]
            tokens = self.tokens.get(base_scope)
            if not tokens or scope not in getattr(tokens, "dependent_scopes", set()):
                return None
            else:
                return tokens
        return self.tokens.get(scope)

    def load_tokens(self):
        """
        May raise an EnvironmentError if the cache file exists but can't be read.
        """
        try:
            with open(self.token_store) as f:
                contents = json.load(f)
                self.tokens = {k: TokenSet(**v) for k, v in contents.items()}
        except FileNotFoundError:
            pass
        except JSONDecodeError:
            raise EnvironmentError(
                "Token cache for Timer CLI is corrupted; please run a `session revoke`"
                " and try again"
            )

    def save_tokens(self):
        def default(x):
            if isinstance(x, set):
                return list(x)
            return str(x)

        if self.modified:
            with open(self.token_store, "w") as f:
                json.dump(
                    {k: v._asdict() for k, v in self.tokens.items()},
                    f,
                    indent=2,
                    sort_keys=True,
                    default=default,
                )
        self.modified = False

    def update_from_oauth_token_response(
        self, token_response: OAuthTokenResponse, original_scopes: Set[str]
    ) -> Dict[str, TokenSet]:
        by_scopes = token_response.by_scopes
        token_sets: Dict[str, TokenSet] = {}
        for scope in by_scopes:
            token_info = by_scopes[scope]
            dependent_scopes = set(s for s in original_scopes if "[" in s)
            token_set = TokenSet(
                access_token=token_info.get("access_token"),
                refresh_token=token_info.get("refresh_token"),
                expiration_time=token_info.get("expires_at_seconds"),
                dependent_scopes=dependent_scopes,
            )
            self.set_tokens(scope, token_set)
            token_sets[scope] = token_set
        self.save_tokens()
        return token_sets


def _get_globus_sdk_native_client(
    client_id: str = CLIENT_ID,
    client_name: str = CLIENT_NAME,
):
    return NativeAppAuthClient(client_id, app_name=client_name)


def safeprint(s):
    try:
        print(s)
        sys.stdout.flush()
    except IOError:
        pass


def _do_login_for_scopes(
    native_client: NativeAppAuthClient, scopes: List[str]
) -> OAuthTokenResponse:
    label = CLIENT_NAME
    host = platform.node()
    if host:
        label = label + f" on {host}"
    native_client.oauth2_start_flow(
        requested_scopes=scopes,
        refresh_tokens=True,
        prefill_named_grant=label,
    )
    linkprompt = "Please log into Globus here"
    safeprint(
        "{0}:\n{1}\n{2}\n{1}\n".format(
            linkprompt, "-" * len(linkprompt), native_client.oauth2_get_authorize_url()
        )
    )
    auth_code = click.prompt("Enter the resulting Authorization Code here").strip()
    return native_client.oauth2_exchange_code_for_tokens(auth_code)


def get_authorizers_for_scopes(
    scopes: List[str],
    token_store: Optional[str] = None,
    client_id: str = CLIENT_ID,
    client_name: str = CLIENT_NAME,
    no_login: bool = False,
) -> Dict[str, GlobusAuthorizer]:
    token_store = token_store or str(DEFAULT_TOKEN_FILE)
    token_cache = TokenCache(token_store)
    token_cache.load_tokens()
    token_sets: Dict[str, TokenSet] = {}
    needed_scopes: Set[str] = set()
    native_client = _get_globus_sdk_native_client(client_id, client_name)

    for scope in scopes:
        token_set = token_cache.get_tokens(scope)
        if token_set is not None:
            token_sets[scope] = token_set
        else:
            needed_scopes.add(scope)

    if len(needed_scopes) > 0 and not no_login:
        token_response = _do_login_for_scopes(native_client, list(needed_scopes))
        new_tokens = token_cache.update_from_oauth_token_response(
            token_response, set(scopes)
        )
        token_sets.update(new_tokens)

    authorizers: Dict[str, GlobusAuthorizer] = {}
    for scope, token_set in token_sets.items():
        if token_set is not None:
            if token_set.refresh_token is not None:

                def refresh_handler(
                    grant_response: OAuthTokenResponse, *args, **kwargs
                ):
                    new_tokens = token_cache.update_from_oauth_token_response(
                        grant_response, set([scope])
                    )

                authorizer = RefreshTokenAuthorizer(
                    token_set.refresh_token,
                    native_client,
                    access_token=token_set.access_token,
                    expires_at=token_set.expiration_time,
                    on_refresh=refresh_handler,
                )
                # Force check that the token is not expired
                authorizer.check_expiration_time()
            else:
                authorizer = AccessTokenAuthorizer(token_set.access_token)
            authorizers[_get_base_scope(scope)] = authorizer
    return authorizers


def get_access_token_for_scope(scope: str) -> Optional[str]:
    authorizer = get_authorizers_for_scopes([scope]).get(_get_base_scope(scope))
    if not authorizer:
        click.echo(f"couldn't obtain authorizer for scope: {scope}", err=True)
        return None
    token = getattr(authorizer, "access_token", None)
    if not token:
        click.echo("authorizer failed to get token from Globus Auth")
        return None
    return token


def logout(token_store: str = DEFAULT_TOKEN_FILE) -> bool:
    try:
        os.remove(token_store)
    except OSError:
        click.echo("couldn't remove token cache file", err=True)
        return False
    return True


def revoke_login(token_store: str = DEFAULT_TOKEN_FILE) -> bool:
    client = _get_globus_sdk_native_client(CLIENT_ID, CLIENT_NAME)
    if not client:
        click.echo("failed to get auth client", err=True)
        return False
    cache = TokenCache(token_store)
    for token_set in cache.tokens.values():
        client.oauth2_revoke_token(token_set.access_token)
        client.oauth2_revoke_token(token_set.refresh_token)
    if not logout(token_store):
        return False
    return True


def get_current_user(
    no_login: bool = False, token_store: str = DEFAULT_TOKEN_FILE
) -> Optional[Dict[str, Any]]:
    """
    When `no_login` is set, returns `None` if not logged in.
    """
    # We don't really care which scope from the AUTH_SCOPE list we use here since they
    # all share the same resource server (Auth itself) and therefore an authorizer for
    # any of them grants us access to the same resource server.
    authorizers = get_authorizers_for_scopes(
        AUTH_SCOPES, token_store=token_store, no_login=no_login
    )
    if not authorizers:
        return None
    auth_client = AuthClient(authorizer=authorizers.get("openid"))
    try:
        user_info = auth_client.oauth2_userinfo()
    except AuthAPIError as e:
        click.echo(
            (
                "Couldn't get user information from Auth service\n"
                "(If you rescinded your consents in the Auth service, do `session"
                " logout` and try again)\n"
                f"    Error details: {str(e)}"
            ),
            err=True,
        )
        sys.exit(1)
    return user_info.data



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['timer_cli']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.1.2,<8.0.0',
 'globus-sdk>=2.0.1,<3.0.0',
 'importlib-metadata>=3.10.0,<4.0.0',
 'requests>=2.24.0,<3.0.0']

entry_points = \
{'console_scripts': ['globus-timer = timer_cli.main:main']}

setup_kwargs = {
    'name': 'globus-timer-cli',
    'version': '0.2.3',
    'description': 'CLI for interacting with the timer API',
    'long_description': '# Globus Automate Timer CLI\n\nThis is a beta-version CLI for use with the (also in beta) Globus Timer API, to\nuse primarily for scheduling recurring Globus transfers through Globus Automate.\n\nAs the CLI and service are still in beta, please feel free to email support at\nglobus dot org with feedback or to resolve issues.\n\n## What is this Service/CLI for?\n\nThe Globus Timer service can be used to schedule recurring transfer tasks. For\nexample, let\'s say we want to have a transfer automatically run every night to\nback up data. We submit a job to the Timer Service starting tonight, and with an\ninterval of 1 day at which it will be re-run. In that request we provide the\nTimer service the same input we would give to the [Globus Transfer Action\nProvider](https://globus-automate-client.readthedocs.io/en/latest/globus_action_providers.html#globus-transfer-transfer-data);\nthat part of the request contains the information for what Globus endpoints we\ntransfer to and from as well as other options relevant to the transfer.\n\n## Installation\n\nThis CLI requires Python version 3.5 or higher. See the [Globus CLI\ndocs](https://docs.globus.org/cli/installation/prereqs/) for help on how to set\nup Python.\n\nOnce the appropriate version of Python is ready, install with `pip\ninstall globus-timer-cli`. This will create a new command line\nutility, `globus-timer` which can be used for all interactions with\nthe service. Online documentation is always available using the\n`--help` option on any command. So, `globus-timer --help` will provide\ninformation about the options on the command while `globus-timer job\n--help` will provide help text specific to working with jobs. When in\ndoubt, add `--help` to a command for guidance on next steps.\n\n## Getting Started\n\nTo schedule transfers on your behalf, this CLI requires authentication through\nGlobus Auth. Initially, the CLI will prompt you to authorize use via the Globus\nAuth system, with a link to open in a web browser. The page will request that\nyou login to your Globus identity and consent to the service looking up your\nidentity and interacting with the Globus Transfer service on your behalf, and\nthen give you an authorization code to copy and paste back into the CLI. Some\nGlobus endpoints require additional authentication for use; when this is\nnecessary, the CLI will repeat the above process to get consent for using the\nGlobus Transfer service and the specific endpoint(s) for that job.\n\nAuthentication information is thereafter cached in the file\n`~/.globus_timer_tokens.json` (so the authentication process is only needed on\nthe first use of the CLI, or of a given transfer endpoint when required); keep\nthis file secret.\n\nAfter first use, you can determine what Globus Auth identity is being used for\ninteracting with the service by running the command:\n\n```\nglobus-timer session whoami\n```\n\nThis will show the identity which will be used for running the Transfers. Be\nsure that this identity is authorized to work with the endpoints involved in the\nTransfer.\n\nTo remove your stored identity information so that you may re-authenticate, for\nexample to invoke subsequent operations using a different Globus Auth identity,\nuse the command:\n\n```\nglobus-timer session logout\n```\n\nNote that should you wish to both logout, and revoke the Timer service\'s\npermission to run further operations, you may use the command `globus-timer\nsession revoke` however this should be an extreme measure as it is much\npreferred to properly delete jobs as described below.\n\n## Scheduling Periodic Globus Transfer operations\n\nNOTE: To avoid confusion, please read the entirety of this section before using\nthe `job transfer` sub-command.\n\nThe starting point for working with the Timer service is scheduling transfers to\nrun periodically. The form of this command is below:\n\n```\nglobus-timer job transfer \\\n    --name example-job \\\n    --label "Timer Transfer Job" \\\n    --interval 28800 \\\n    --start \'2020-01-01T12:34:56\' \\\n    --source-endpoint ddb59aef-6d04-11e5-ba46-22000b92c6ec \\\n    --dest-endpoint ddb59af0-6d04-11e5-ba46-22000b92c6ec \\\n    --item ~/file1.txt ~/new_file1.txt false \\\n    --item ~/file2.txt ~/new_file2.txt false\n```\n\nThe parameters to the `job transfer` command are as follows (also documented in\nthe CLI help: `globus-timer job transfer --help`):\n\n* `name`: A friendly name you can use to identify the job. However,\n  note that all operations on the job (see below) are based on the job\n  identifier (`job_id`) which is returned in the output of this\n  command.\n\n* `label`: A label string which will be used on the Globus Transfer\n  tasks which will be created to perform the periodic transfer\n  operations. You may use this to help you identify the source of the\n  Globus Transfer task when viewing them in the Globus web\n  application. If no `label` string is provided, the Globus Transfer\n  label will include the `name` of this Timer job.\n\n* `interval`: The time, in seconds, between invocations of the Globus\n  Transfer operation. The Transfer will be invoked after this number\n  of fixed seconds. Note that this easily allows for intervals such as\n  hourly (3600 seconds), daily (86400 seconds) or weekly (604800\n  seconds), but it doesn\'t allow for day of month or provide the\n  ability to compensate for other time related changes like daylight\n  savings time changes. This is presently a limitation of the Timer\n  service. In some cases, the actual time the operation is started in\n  Globus Transfer may be a few seconds longer than the specified\n  interval, but any delay will not impact scheduling of the next\n  operation (i.e. the job will stay on schedule and not "drift" due to\n  delays in any particular execution).\n\n* `start`: A timestamp defining when the job should first be\n  scheduled. Thus, the first invocation of the Globus Transfer\n  operation will occur at the specified time and then again after each\n  `interval` number of seconds. The time may contain just a year,\n  month and date, or may also contain the exact time of day as shown\n  in the example. Unless a timezone is specified, the local timezone\n  will be used.\n\n* `source-endpoint` and `dest-endpoint`: Are the ids for Globus\n  Transfer endpoints for the source and destination of the transfer\n  operation respectively. These values can be retrieved from the\n  Globus web application.\n\n* `item`: Specifies the exact locations on the source and destination\n  endpoints to perform the transfer operation. `item` should be\n  followed by three values: the path on the source endpoint from which\n  to start the transfer, the path on the destination endpoint where\n  the data should be placed, and a true or false value indicating\n  whether the transfer should be performed in a recursive manner. When\n  the value is true, if the source path names a folder, the folder and\n  all of its contained files and other folders will be transferred. If\n  the value is false, the first (source) path should refer to a\n  specific file.\n\nNote that, as shown in the example, a single job creation command can\ncontain multiple specifications of the `item` parameter. This means\nthat a single job can transfer from multiple paths between a single\nsource endpoint and a single destination endpoint.\n\nAs an alternative to specifying the `item` parameter one or more\ntimes, the item values may be stored in a file in which each line has\nthe same format as the `item` parameter. The file may be specified\nwith the `items-file` option which is used *instead of* the `item`\nparameter. An example of the file which would perform the same\noperations as the example above would look like:\n\n```\n# This is my items file\n~/file1.txt, ~/new_file1.txt, false\n~/file2.txt, ~/new_file2.txt, false\n```\n\nIf this file was named `transfer_items.csv` it would be specified with the parameter:\n\n```\n--items-file transfer_items.csv\n```\n\nNote that this is a CSV file. The individual parts of the item are separated\nwith commas, and lines starting with a `#` are considered to be comments and are\nnot processed as items for the transfer job.\n\n### Additional Transfer Options\n\nSome optional flags for setting certain Transfer options include:\n\n* `--sync-level`, adjusting how Transfer should behave if files already exist at\n  the destination\n* `--verify-checksum`, causing Transfer to retry if file checksums don\'t match\n* `--encrypt-data`, causing Transfer to encrypt data using TLS\n* `--preserve-timestamp`, causing Transfer to keep file timestamps the same at\n  the destination\n\nYou may also specify a finite number of times or duration after which a job\nshould stop:\n\n* `--stop-after-runs` takes a number, telling Timer to run the job only that\n  many times.\n* `--stop-after-date` takes a date argument, telling Timer to continue running\n  the job only until the given date.\n\nWhen either condition is reached, your job will still exist, and results will\nstill be visible, but Timer will not run any more transfers. If you specify\nboth, the job stops running as soon as one stop condition is met. If you update\nan existing job to modify either of these, Timer will resume running the job\nuntil it hits your new stop condition.\n\nAs always, all of the above options are also documented in `globus-timer job\ntransfer --help`.\n\n## Monitoring and Controlling Submitted jobs\n\nAfter submitting the transfer job, the CLI should return some results containing\na `job_id`, which identifies this job in the Timer service. To check the\nstatus of your jobs, use:\n\n```\nglobus-timer job status <job_id>\n```\n\nThis command defaults to a summarized version of the job\'s information, which\ndoes not include the full details for the corresponding task in Globus Transfer. To\ncheck those, use `-v/--verbose`:\n\n```\nglobus-timer job status --verbose JOB_ID\n```\n\nCommands return date-times in ISO format, in UTC time, so the timezone\nmost likely will not match the local time where you run the\ncommand. No need to worry: the actual start time is still equal to\nyour submission\'s start time, etc.\n\nThe job description will also contain a value for `status` indicating\nthe condition of the job. The `status` may contain one of the\nfollowing values:\n\n* `new`: The job has been received, but it has not yet been scheduled\n  for its first execution (for example the start time has not yet been\n  reached).\n\n* `loaded`: The job has reached its execution mode and is executing\n  repeatedly on the requested interval. This is the "steady-state" for\n  a job when it is executing as expected.\n\n* `updated`: If the job has recently been updated, it will be in this\n  state until the job has transitioned to its new regular\n  operation. This state should rarely be encountered as it is a\n  transient state immediately after a job update.\n\n* `deleted`: Indicates that the job has been deleted. This would not\n  normally be seen, but the `job status` command described above as\n  well as other commands described below provide the option of\n  retrieving jobs marked as deleted using a command line flag\n  `--show-deleted`.\n\nA final important note: `Last Result` in the non-verbose output extends only as\nfar as the Globus Automate system: `SUCCESS` indicates you have successfully submitted\nyour job to the Timer service, which in turn successfully sent the task to the\nGlobus Transfer Action Provider. It\'s possible that the Globus Transfer service will\nsubsequently encounter some error running your transfer. Check the `--verbose`\noutput, which includes the actual response from Transfer, to be certain that\nTransfer has run your job successfully.\n\nTo see all of your jobs which are outstanding, use the command:\n\n```\nglobus-timer job list\n```\n\nThis is particularly helpful if you do not have the `job_id` for a\nparticular job available. It will show all of your jobs in the same\nformat as the `job status` command described above. Similarly, the\n`--verbose/-v` and `--show-deleted` options are supported.\n\nFinally, a job may be removed from operation using the command:\n\n```\nglobus-timer job delete <job_id>\n```\n\nThe job will be removed from operation and no further transfers will\nbe performed. The same information provided via `job status` will be\ndisplayed, showing the final state of the job. As described above, the\n`job status` command may still retrieve the job information by\nproviding the `--show-deleted` flag. This can be helpful to review the\nfinal state of the job, such as the last transfer or error condition,\neven if the output for the `job delete` command is lost.\n',
    'author': 'Rudyard Richter',
    'author_email': 'rudyard@globus.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
